/******************************************************************************
  (c) Copyright 2002-2006, 1060 Research Ltd                                   

  This Software is licensed to You, the licensee, for use under the terms of   
  the 1060 Public License v1.0. Please read and agree to the 1060 Public       
  License v1.0 [www.1060research.com/license] before using or redistributing   
  this software.                                                               

  In summary the 1060 Public license has the following conditions.             
  A. You may use the Software free of charge provided you agree to the terms   
  laid out in the 1060 Public License v1.0                                     
  B. You are only permitted to use the Software with components or applications
  that provide you with OSI Certified Open Source Code [www.opensource.org], or
  for which licensing has been approved by 1060 Research Limited.              
  You may write your own software for execution by this Software provided any  
  distribution of your software with this Software complies with terms set out 
  in section 2 of the 1060 Public License v1.0                                 
  C. You may redistribute the Software provided you comply with the terms of   
  the 1060 Public License v1.0 and that no warranty is implied or given.       
  D. If you find you are unable to comply with this license you may seek to    
  obtain an alternative license from 1060 Research Limited by contacting       
  license@1060research.com or by visiting www.1060research.com                 

  NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC  
  LICENSE V1.0 FOR DETAILS                                                     

  THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ     
  THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license      

  File:          $RCSfile: HTTPRequestSpace.java,v $
  Version:       $Name:  $ $Revision: 1.11 $
  Last Modified: $Date: 2008/05/20 10:45:52 $
 *****************************************************************************/
package org.netkernel.http.transport;

import java.io.*;
import java.net.URI;
import java.util.*;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;

import org.netkernel.http.transport.representation.ParameterUploadAspect;
import org.netkernel.http.util.MultiPartRequest;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.NKFSpaceImpl;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.Utils;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.urii.IEndpoint;


/** Context and Endpoint to fulfill requests for bridge provided httpRequest resources
 * @author pjr
 *
 */
public class HTTPRequestSpace extends NKFSpaceImpl
{
	public static final String HTTP_REQUEST_SCHEME="httpRequest:/";
	public static final String HTTP_RESPONSE_SCHEME="httpResponse:/";
	public static final String HTTP_PARAM_SET="params";
	public static final String HTTP_PARAM_SPACE="param/";
	public static final String HTTP_URL="url";
	public static final String HTTP_BODY="body";
	public static final String HTTP_UPLOAD="upload/";
	public static final String HTTP_METHOD="method";
	public static final String HTTP_HEADER="header/";
	public static final String HTTP_HEADERS="headers";
	public static final String HTTP_RESPONSE_CODE="code";
	public static final String HTTP_COOKIE="cookie/";
	public static final String HTTP_COOKIE_SINK="cookie";
	public static final String HTTP_REMOTEHOST="remote-host";
	public static final String HTTP_REDIRECT="redirect";
	public static final String HTTP_LOW_LEVEL_REQUEST="advanced/HttpServletRequest";
	
	private final HttpServletRequest mRequest;
	private final HTTPBridgeOverlayConfig mConfig;
	private static IEndpoint sEndpoint = new HTTPRequestEndpoint();
	
	/** map of String -> IReadableBinaryStreamRepresentation */
	private Map mUploads;
	/** content of request body (post or put of data) */
	private IReadableBinaryStreamRepresentation mContent;
	/** parameters */
	private IHDSNode mParam;
	
	//Response Resources
	private Map mResponseCookies=new HashMap(2);
	private Map mResponseHeaders=new HashMap(2);
	private int mResponseCode=200;
	private String mResponseRedirect;

	public HTTPRequestSpace(HttpServletRequest aRequest, HTTPBridgeOverlayConfig aConfig )
	{	mRequest = aRequest;
		mConfig=aConfig;
	}
	
	public String toString()
	{	return "HTTPRequestSpace";
	}

	public void onResolve(INKFResolutionContext helper) throws Exception
	{	String requestId = helper.getRequestToResolve().getIdentifier();
		if (helper.getRequestToResolve().getVerb()!=INKFRequestReadOnly.VERB_TRANSREPRESENT && (requestId.startsWith(HTTP_REQUEST_SCHEME) || requestId.startsWith(HTTP_RESPONSE_SCHEME)))
		{	sEndpoint.onCommissionEndpoint(getKernel(), this);
			helper.createResolutionResponse(sEndpoint);
		}
	}

	public void handleRequestFromEndpoint(INKFRequestContext helper) throws Exception
	{	Object response=null;
		String contentType="text/plain";
		int verb=helper.getThisRequest().getVerb();
		String identifier=helper.getThisRequest().getIdentifier();
	
		if(identifier.startsWith(HTTP_REQUEST_SCHEME))
		{	if(verb!=INKFRequestReadOnly.VERB_SOURCE && verb!=INKFRequestReadOnly.VERB_EXISTS)
			{	throw new NKFException("httpRequest is SOURCE,EXISTS only");				
			}
			String type = identifier.substring(HTTP_REQUEST_SCHEME.length());
			if (type.equals(HTTP_PARAM_SET))
			{	ensureParamsProcessed();
				response=mParam;
				contentType=null;
			}
			else if (type.startsWith(HTTP_PARAM_SPACE))
			{	ensureParamsProcessed();
				String param = type.substring(HTTP_PARAM_SPACE.length());
				response=mParam.getFirstValue(param);
				contentType=null;
			}
			else if (type.equals(HTTP_BODY))
			{	ensureParamsProcessed();
				if (mContent!=null)
				{	response= mContent;
					contentType=mRequest.getContentType();
				}
			}
			else if (type.startsWith(HTTP_URL))
			{	response = mRequest.getRequestURL().toString();
			}
			else if (type.startsWith(HTTP_UPLOAD))
			{	ensureParamsProcessed();
				String upload = type.substring(HTTP_UPLOAD.length());
				if(mUploads==null)
				{	throw new Exception("No file uploads - check form enctype attribute is \"multipart/form-data\" ");					
				}
				response = mUploads.get(upload);
				contentType=null;
			}
			else if (type.equals(HTTP_METHOD))
			{	helper.createResponseFrom(mRequest.getMethod());
			}
			else if (type.startsWith(HTTP_HEADER))
			{	String field = type.substring(HTTP_HEADER.length());
				response = mRequest.getHeader(field);
			}
			else if (type.startsWith(HTTP_HEADERS))
			{	HDSBuilder b=new HDSBuilder();
				Enumeration e=mRequest.getHeaderNames();
				while(e.hasMoreElements())
				{	String name=(String)e.nextElement();
					Enumeration f=mRequest.getHeaders(name);
					while(f.hasMoreElements())
					{	String value=(String)f.nextElement();
						b.addNode(name, value);						
					}
				}
				response = b.getRoot();
			}
			else if (type.startsWith(HTTP_COOKIE))
			{	String name = type.substring(HTTP_COOKIE.length());
				Cookie[] cookies = mRequest.getCookies();
				if(cookies!=null)
				{	for(int i=0; i<cookies.length; i++)
					{	if (cookies[i].getName().equals(name))
						{	response = cookies[i];
							break;
						}
					}
				}
			}
			else if (type.equals(HTTP_REMOTEHOST))
			{	response=mRequest.getRemoteHost();
			}
			else if (type.equals(HTTP_LOW_LEVEL_REQUEST))
			{	response=mRequest;
			}
			else
			{	INKFRequestReadOnly req = helper.getThisRequest();
				String verb2=RequestFactory.verbString(req.getVerb());
				NKFException e=helper.createFormattedException("EX_BAD_HTTP_RESOURCE", "MSG_BAD_HTTP_RESOURCE", null,null,verb2,req.getIdentifier());
				throw e;
			}
			
			if (helper.getThisRequest().getVerb()==INKFRequestReadOnly.VERB_EXISTS)
			{	boolean exists = response!=null;
				helper.createResponseFrom( exists ? Boolean.TRUE : Boolean.FALSE );
			}
			else
			{	if (response!=null)
				{	INKFResponse nkfResp=helper.createResponseFrom(response);
					nkfResp.setExpiry(INKFResponse.EXPIRY_NEVER);
					if (contentType!=null)
					{	nkfResp.setMimeType(contentType);
					}
				}
			}
		}
		else
		{	String type = identifier.substring(HTTP_RESPONSE_SCHEME.length());
			if(verb!=INKFRequestReadOnly.VERB_SINK)
			{	throw new NKFException("httpResponse is SINK only");
			}
			Object value=helper.getThisRequest().getPrimary();
			innerProcessResponseSink(type,value,helper);
		}
	}
	
	private void innerProcessResponseSink(String aType, Object aValue, INKFRequestContext aHelper) throws NKFException
	{
		if (aType.startsWith(HTTP_COOKIE_SINK))
		{	Cookie cookie=transreptIfNeeded(aValue,Cookie.class,aHelper);
			mResponseCookies.put(cookie.getName()+cookie.getPath(), cookie);
		}
		else if(aType.startsWith(HTTP_HEADER))
		{	String name = aType.substring(HTTP_HEADER.length());
			//String value=aValue.toString(); //transreptIfNeeded(aValue,String.class,aHelper);
			mResponseHeaders.put(name, aValue);		
		}
		else if(aType.startsWith(HTTP_RESPONSE_CODE))
		{	Integer cd=transreptIfNeeded(aValue,Integer.class,aHelper);
			mResponseCode=cd.intValue();
		}
		else if(aType.startsWith(HTTP_REDIRECT))
		{	String redirectURIString=transreptIfNeeded(aValue,String.class,aHelper);
			int i=redirectURIString.indexOf(':');
			if (i>=0)
			{	URI uri=URI.create(redirectURIString);
				String scheme=uri.getScheme();
				if (scheme!=null && !scheme.equals("http") && !scheme.equals("https"))
				{	//remove scheme
					redirectURIString=uri.getSchemeSpecificPart();
				}
			}
			mResponseRedirect=redirectURIString;
		}
	}
	
	private <T> T transreptIfNeeded(Object aRep, Class<T> aTo, INKFRequestContext aHelper) throws NKFException
	{	Object result=aRep;
		if (!aTo.isAssignableFrom(aRep.getClass()))
		{	result=aHelper.transrept(aRep, aTo);
		}
		return (T)result;
	}
	
	
	private void ensureParamsProcessed() throws IOException
	{	
		String encoding=mRequest.getCharacterEncoding();
		if (mParam==null)
		{	String mime=mRequest.getContentType();
	
			HDSBuilder b = new HDSBuilder();
			
			if (mime!=null && mime.startsWith("multipart/form-data"))
			{	// multipart content = uploads
				MultiPartRequest mpr=new MultiPartRequest(mRequest);
				String[] names=mpr.getPartNames();
				String localEncoding=encoding;
				if(localEncoding==null)
				{	localEncoding=mConfig.getDefaultPostParamEncoding();
				}
				for(int i=0; i<names.length;i++)
				{	String name=names[i];
					String filename=mpr.getFilename(name);
					if(filename!=null)
					{	b.addNode(name,filename);
						if (mUploads==null)
						{	mUploads = new HashMap();
						}
						mUploads.put(name,new ParameterUploadAspect(mpr,name));
					}
					else
					{	// must be form data in addition to the upload
						InputStream is=mpr.getInputStream(name);
						InputStreamReader isr=new InputStreamReader(is, localEncoding); 	//Maybe should try to get the encoding from the MPR - but it has no access - so we use the mRequest's value above
						StringWriter sw=new StringWriter(mpr.getContentLength(name));
						while(isr.ready())
						{	sw.write(isr.read());
						}
						b.addNode(name,sw.toString());
					}
				}
			}
			
			if (mime==null || (mime!=null && !mime.startsWith("application/x-www-form-urlencoded") && !mime.startsWith("multipart/form-data")))
			{	// special body content
				
				if (mRequest.getContentLength()>0)
				{	ByteArrayOutputStream baos = new ByteArrayOutputStream(mRequest.getContentLength());
					Utils.pipe(mRequest.getInputStream(), baos);
					mContent = new ByteArrayRepresentation(baos, mRequest.getCharacterEncoding());
				}
			}
			
			// fix to work around unknown encoding of post parameters
			// i.e. the standard says they should be same as the encoding of the HTML form that contains the FORM
			// - not very useful
			final String JETTY_ASSUMED_ENCODING="ISO-8859-1";
			String assumedEncoding=mConfig.getDefaultPostParamEncoding();
			boolean unmanglePostParams= (mRequest.getMethod().equals("POST") && encoding==null  &&  !assumedEncoding.equals(JETTY_ASSUMED_ENCODING));
			
			Enumeration i=mRequest.getParameterNames();
			while(i.hasMoreElements())
			{	String name=(String)i.nextElement();
				for (String value: mRequest.getParameterValues(name) )
				{	if (unmanglePostParams)
					{	byte[] b2=value.getBytes(JETTY_ASSUMED_ENCODING);
						value=new String(b2,assumedEncoding);
					}
					b.addNode(name, value);
				}
			}
			mParam = b.getRoot();

		}
	}
	
	public boolean resolveEquals(Object aOther)
	{	return aOther instanceof HTTPRequestSpace;
	}
	
	public Map getResponseCookies()
	{	return mResponseCookies;
	}
	public Map getResponseHeaders()
	{	return mResponseHeaders;
	}
	public int getResponseCode()
	{	return mResponseCode;		
	}
	public String getResponseRedirect()
	{	return mResponseRedirect;		
	}
	public boolean shouldRedirect()
	{	return mResponseRedirect!=null;
	}
	
	void applyResponseMetadata(IRequestResponseFields aResponseMetadata, INKFRequestContext aContext) throws NKFException
	{	for (int i=0; i<aResponseMetadata.size(); i++)
		{	String key=aResponseMetadata.getKey(i);
			if (key.startsWith(HTTP_RESPONSE_SCHEME))
			{	key=key.substring(HTTP_RESPONSE_SCHEME.length());
				Object value=aResponseMetadata.getValue(i);
				innerProcessResponseSink(key,value,aContext);
			}
		}
	}
}