package org.netkernel.http.transport;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.URLDecoder;
import java.util.concurrent.atomic.AtomicInteger;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.mortbay.jetty.HttpHeaders;
import org.mortbay.jetty.handler.AbstractHandler;
import org.netkernel.http.transport.representation.HTTPRequestResponseRepresentation;
import org.netkernel.layer0.nkf.INKFLocale;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.urii.INetKernelThrowable;


public class NetKernelHandler extends AbstractHandler implements INetKernelHandler 
{
	private INKFRequestContext mContext;
	private AtomicInteger mRequestCount=new AtomicInteger();
	
	public NetKernelHandler()
	{
	}
	
	public void destroy()
	{	super.destroy();
		mContext=null;
	}
	
	public void registerContext(INKFRequestContext aHelper)
	{	mContext=aHelper;
 		mContext.logFormatted(INKFLocale.LEVEL_INFO,"MSG_REGISTERED_SYNC_HANDLER", new Object[] {null});
	}

	public void handle(String target, HttpServletRequest request,	HttpServletResponse response, int dispatch) throws IOException, ServletException
	{	try
		{	mRequestCount.incrementAndGet();
			HTTPRequestResponseRepresentation rep = new HTTPRequestResponseRepresentation(request,response);
			String url=URLDecoder.decode(request.getRequestURL().toString(), "UTF-8");
			INKFRequest req = mContext.createRequest(url);
			req.addPrimaryArgument(rep);
			mContext.issueRequestForResponse(req);
		}
		catch (Throwable t)
		{	response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
			response.setDateHeader(HttpHeaders.EXPIRES,0);
			response.setContentType("text/plain");
			response.setHeader(HttpHeaders.SERVER, HTTPTransportEndpoint.getServerHeader() );
			OutputStream os=response.getOutputStream();
			PrintWriter pw = new PrintWriter(os);
			if (t instanceof INetKernelThrowable)
			{	pw.write(t.toString());
			}
			else
			{	t.printStackTrace(pw);
			}
			pw.flush();
		}
		finally
		{	response.flushBuffer();
		}	
	}

	public int requestCount()
	{	return mRequestCount.get();
	}

}
