package org.netkernel.http.transport;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.concurrent.atomic.AtomicInteger;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.mortbay.jetty.HttpConnection;
import org.mortbay.jetty.HttpHeaders;
import org.mortbay.jetty.Request;
import org.mortbay.jetty.handler.ContextHandler;
import org.mortbay.jetty.security.HTAccessHandler;
import org.mortbay.resource.Resource;
import org.netkernel.http.transport.representation.HTTPRequestResponseRepresentation;
import org.netkernel.layer0.nkf.INKFLocale;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.urii.INetKernelThrowable;


public class NetKernelHandlerHTAccess extends HTAccessHandler implements INetKernelHandler 
{
	private INKFRequestContext mContext;
	private AtomicInteger mRequestCount=new AtomicInteger();
	private ContextHandler mLocalContextHandler=new LocalContextHandler();
	
	public NetKernelHandlerHTAccess()
	{	this.setAccessFile("htaccess");
	}
	
	public void destroy()
	{	super.destroy();
		mContext=null;
	}
	
	public void registerContext(INKFRequestContext aContext)
	{	mContext=aContext;
 		mContext.logFormatted(INKFLocale.LEVEL_INFO,"MSG_REGISTERED_SYNC_HANDLER", new Object[] {null});
 		super.setProtegee(mLocalContextHandler);
	}

	public void handle(String target, HttpServletRequest request,	HttpServletResponse response, int dispatch) throws IOException, ServletException
	{	super.handle(target, request, response, dispatch);
		Request basereq=(Request)request;
		if(basereq.isHandled()) return;
		try
		{	mRequestCount.incrementAndGet();
			HTTPRequestResponseRepresentation rep = new HTTPRequestResponseRepresentation(request,response);
			INKFRequest req = mContext.createRequest(request.getRequestURL().toString());
			req.addPrimaryArgument(rep);
			mContext.issueRequestForResponse(req);
		}
		catch (Throwable t)
		{	response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
			response.setDateHeader(HttpHeaders.EXPIRES,0);
			response.setContentType("text/plain");
			response.setHeader(HttpHeaders.SERVER, HTTPTransportEndpoint.getServerHeader() );
			OutputStream os=response.getOutputStream();
			PrintWriter pw = new PrintWriter(os);
			if (t instanceof INetKernelThrowable)
			{	pw.write(t.toString());
			}
			else
			{	t.printStackTrace(pw);
			}
			pw.flush();
		}
		finally
		{	response.flushBuffer();
		}	
	}

	public int requestCount()
	{	return mRequestCount.get();
	}

	private class LocalContextHandler extends ContextHandler
	{
		@Override
		public Resource getResource(String res) throws MalformedURLException
		{
			try
			{	URI url=URI.create(HttpConnection.getCurrentConnection().getRequest().getRequestURL().toString());
				StringBuilder sb=new StringBuilder();
				sb.append(url.getScheme());
				sb.append("://");
				sb.append(url.getRawAuthority());
				sb.append(res);
				INKFRequest req = mContext.createRequest(sb.toString());
				req.setRepresentationClass(IReadableBinaryStreamRepresentation.class);
				req.setHeader(HTTPBridgeOverlayConfig.HTTPBRIDGE_TRANSPARENT_PASS_THROUGH, true);
				INKFResponseReadOnly<IReadableBinaryStreamRepresentation> resp=mContext.issueRequestForResponse(req);
				return new LocalResource(resp);
			}
			catch(Exception e)
			{	mContext.logRaw(INKFRequestContext.LEVEL_FINE, e.toString());				
			}
			return null;
		}
	}
	
	private class LocalResource extends Resource
	{
		private INKFResponseReadOnly<IReadableBinaryStreamRepresentation> mResponse;
		private long mLastModified;
		
		public LocalResource(INKFResponseReadOnly<IReadableBinaryStreamRepresentation> aResponse)
		{	mResponse=aResponse;
			mLastModified=0;
		}
		
		@Override
		public Resource addPath(String arg0) throws IOException,
				MalformedURLException
		{
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public boolean delete() throws SecurityException
		{
			// TODO Auto-generated method stub
			return false;
		}

		@Override
		public boolean exists()
		{
			return true;
		}

		@Override
		public File getFile() throws IOException
		{
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public InputStream getInputStream() throws IOException
		{
			return mResponse.getRepresentation().getInputStream();
		}

		@Override
		public String getName()
		{
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public OutputStream getOutputStream() throws IOException,
				SecurityException
		{
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public URL getURL()
		{
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public boolean isDirectory()
		{
			return false;
		}

		@Override
		public long lastModified()
		{
			if(mResponse.isExpired())
			{	mLastModified=1;				
			}
			return mLastModified;
		}

		@Override
		public long length()
		{
			// TODO Auto-generated method stub
			return 0;
		}

		@Override
		public String[] list()
		{
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public void release()
		{
			// TODO Auto-generated method stub
			
		}

		@Override
		public boolean renameTo(Resource arg0) throws SecurityException
		{
			// TODO Auto-generated method stub
			return false;
		}
		
	}
	
}
